struct DirectionalLight {
	float4 Ambient;
	float4 Diffuse;
	float4 Specular;
	float3 Direction;
	float pad;
};

struct PointLight {
	float4 Ambient;
	float4 Diffuse;
	float4 Specular;

	float3 Position;
	float Range;

	float3 Att;
	float pad;
};

struct SpotLight {
	float4 Ambient;
	float4 Diffuse;
	float4 Specular;

	float3 Position;
	float Range;

	float3 Direction;
	float Spot;

	float3 Att;
	float pad;
};

struct Material {
	float4 Ambient;
	float4 Diffuse;
	float4 Specular; // w = SpecPower
	float4 Reflect;
};

void ComputeDirectionalLight(Material mat, DirectionalLight L, float3 normal, float3 toEye, out float4 ambient, out float4 diffuse, out float4 spec) {
	// Initialize outputs.
	ambient = float4(0.0f, 0.0f, 0.0f, 0.0f);
	diffuse = float4(0.0f, 0.0f, 0.0f, 0.0f);
	spec = float4(0.0f, 0.0f, 0.0f, 0.0f);

	// The light vector aims opposite the direction the light rays travel.
	float3 lightVec = -L.Direction;

	// Add ambient term.
	ambient = mat.Ambient * L.Ambient;

	// Add diffuse and specular term, provided the surface is in 
	// the line of site of the light.

	float diffuseFactor = dot(lightVec, normal);

	// Flatten to avoid dynamic branching.
	[flatten]
	if (diffuseFactor > 0.0f) {
		float3 v = reflect(-lightVec, normal);
		float specFactor = pow(max(dot(v, toEye), 0.0f), mat.Specular.w);
		
		//if (diffuseFactor > 0.5f) diffuseFactor = 1.0f;
		//if (diffuseFactor > 0.15f && diffuseFactor <= 0.5f) diffuseFactor = 0.5f;
		//if (diffuseFactor < 0.15f) diffuseFactor = 0.0f;
		
		//if (specFactor > 0.8f) specFactor = 1.0f;
		//if (specFactor > 0.1f && specFactor <= 0.8f) specFactor = 0.5f;
		//if (specFactor < 0.1f) specFactor = 0.0f;

		diffuse = diffuseFactor * mat.Diffuse * L.Diffuse;
		spec = specFactor * mat.Specular * L.Specular;
	}
}

void ComputePointLight(Material mat, PointLight L, float3 pos, float3 normal, float3 toEye, out float4 ambient, out float4 diffuse, out float4 spec) {
	// Initialize outputs.
	ambient = float4(0.0f, 0.0f, 0.0f, 0.0f);
	diffuse = float4(0.0f, 0.0f, 0.0f, 0.0f);
	spec = float4(0.0f, 0.0f, 0.0f, 0.0f);

	// The vector from the surface to the light.
	float3 lightVec = L.Position - pos;

	// The distance from surface to light.
	float d = length(lightVec);

	// Range test.
	if (d > L.Range)
		return;

	// Normalize the light vector.
	lightVec /= d;

	// Ambient term.
	ambient = mat.Ambient * L.Ambient;

	// Add diffuse and specular term, provided the surface is in 
	// the line of site of the light.

	float diffuseFactor = dot(lightVec, normal);

	// Flatten to avoid dynamic branching.
	[flatten]
	if (diffuseFactor > 0.0f) {
		float3 v = reflect(-lightVec, normal);
		float specFactor = pow(max(dot(v, toEye), 0.0f), mat.Specular.w);

		diffuse = diffuseFactor * mat.Diffuse * L.Diffuse;
		spec = specFactor * mat.Specular * L.Specular;
	}

	// Attenuate
	float att = 1.0f / dot(L.Att, float3(1.0f, d, d * d));

	diffuse *= att;
	spec *= att;
}

void ComputeSpotLight(Material mat, SpotLight L, float3 pos, float3 normal, float3 toEye, out float4 ambient, out float4 diffuse, out float4 spec) {
	// Initialize outputs.
	ambient = float4(0.0f, 0.0f, 0.0f, 0.0f);
	diffuse = float4(0.0f, 0.0f, 0.0f, 0.0f);
	spec = float4(0.0f, 0.0f, 0.0f, 0.0f);

	// The vector from the surface to the light.
	float3 lightVec = L.Position - pos;

	// The distance from surface to light.
	float d = length(lightVec);

	// Range test.
	if (d > L.Range)
		return;

	// Normalize the light vector.
	lightVec /= d;

	// Ambient term.
	ambient = mat.Ambient * L.Ambient;

	// Add diffuse and specular term, provided the surface is in 
	// the line of site of the light.

	float diffuseFactor = dot(lightVec, normal);

	// Flatten to avoid dynamic branching.
	[flatten]
	if (diffuseFactor > 0.0f) {
		float3 v = reflect(-lightVec, normal);
		float specFactor = pow(max(dot(v, toEye), 0.0f), mat.Specular.w);

		diffuse = diffuseFactor * mat.Diffuse * L.Diffuse;
		spec = specFactor * mat.Specular * L.Specular;
	}

	// Scale by spotlight factor and attenuate.
	float spot = pow(max(dot(-lightVec, L.Direction), 0.0f), L.Spot);

	// Scale by spotlight factor and attenuate.
	float att = spot / dot(L.Att, float3(1.0f, d, d * d));

	ambient *= spot;
	diffuse *= att;
	spec *= att;
}

float3 NormalSampleToWorldSpace(float3 normalMapSample, float3 unitNormalW, float3 tangentW) {
	// Uncompress each component from [0,1] to [-1,1].
	float3 normalT = 2.0f * normalMapSample - 1.0f;

	// Build orthonormal basis.
	float3 N = unitNormalW;
	float3 T = normalize(tangentW - dot(tangentW, N) * N);
	float3 B = cross(N, T);

	float3x3 TBN = float3x3(T, B, N);

	// Transform from tangent space to world space.
	float3 bumpedNormalW = mul(normalT, TBN);

	return bumpedNormalW;
}

Texture2D shaderTexture         : register(t0);
//Texture2D shaderSpecularTexture : register(t1);
//Texture2D shaderNormalTexture	: register(t2);

cbuffer MatrixBuffer : register(b0) {
	matrix worldMatrix;
	matrix viewMatrix;
	matrix projectionMatrix;
	matrix worldInvTranspose; // for converting normals
	Material gMaterial;
};

cbuffer cbPerFrame : register(b1) {
	DirectionalLight	gDirLight;
	SpotLight			gSpotLight;
	float3				gEyePosW;
	float				pad;
};

SamplerState SampleType;

struct PixelInputType {
	float4 position	: SV_POSITION;
	float2 tex		: TEXCOORD;
	float3 PosW     : POSITION;
	float3 NormalW  : NORMAL;
	float3 TangentW : TANGENT;
};

float4 PixelShaderFX(PixelInputType input) : SV_TARGET {

	//float4 textureColor;

	// Sample the pixel color from the texture using the sampler at this texture coordinate location.
	//textureColor = shaderTexture.Sample(SampleType, input.tex);

	//return textureColor;

	// Interpolating normal can unnormalize it, so normalize it.
	input.NormalW = normalize(input.NormalW);

	float3 toEye = gEyePosW - input.PosW;
	float3 toEyeW = normalize(gEyePosW - input.PosW);
	
	// Cache the distance to the eye from this surface point.
	float distToEye = length(toEye);

	Material mt = gMaterial;
	////--mt.Ambient = float4(0.1f, 0.1f, 0.1f, 1.0f);
	////mt.Diffuse = float4(0.9f, 0.9f, 0.9f, 1.0f);
	////--mt.Diffuse = shaderTexture.Sample(SampleType, input.tex);
	mt.Diffuse = shaderTexture.Sample(SampleType, input.tex);
	////mt.Diffuse = shaderSpecularTexture.Sample(SampleType, input.tex);
	////mt.Diffuse = shaderNormalTexture.Sample(SampleType, input.tex);
	////mt.Specular = shaderSpecularTexture.Sample(SampleType, input.tex) + 0.3f;
	////--mt.Specular = shaderTexture.Sample(SampleType, input.tex);
	////mt.Specular = float4(0.8f, 0.8f, 0.8f, 1.0f);
	////--mt.Specular.w = 16.0f;
	////mt.Reflect = float4(0.1f, 0.1f, 0.1f, 1.0f);
	//Material mt;
	//mt.Ambient = float4(0.1f, 0.1f, 0.1f, 1.0f);
	//mt.Ambient = gMaterial.Ambient;
	//mt.Diffuse = shaderTexture.Sample(SampleType, input.tex);
	//mt.Specular = shaderTexture.Sample(SampleType, input.tex);
	//mt.Specular.w = 16.0f;
	//mt.Reflect = float4(0.1f, 0.1f, 0.1f, 1.0f);

	//float3 normalMapSample = shaderNormalTexture.Sample(SampleType, input.tex).rgb;
	//float3 bumpedNormalW = NormalSampleToWorldSpace(normalMapSample, input.NormalW, input.TangentW);

	// Start with a sum of zero. 
	float4 ambient = float4(0.0f, 0.0f, 0.0f, 0.0f);
	float4 diffuse = float4(0.0f, 0.0f, 0.0f, 0.0f);
	float4 spec = float4(0.0f, 0.0f, 0.0f, 0.0f);

	// Sum the light contribution from each light source.
	float4 A, D, S;

	ComputeDirectionalLight(mt, gDirLight, input.NormalW, toEyeW, A, D, S);
	//ComputeDirectionalLight(mt, gDirLight, bumpedNormalW, toEyeW, A, D, S);

	ambient += A;
	diffuse += D;
	spec += S;

	float4 litColor = ambient + diffuse + spec;
	
	float  gFogStart = 0.0f;
	float  gFogRange = 16.0f;
	float4 gFogColor = float4(0.0f, 0.0f, 0.0f, 1.0f);
	float fogLerp = saturate( (distToEye - gFogStart) / gFogRange ); 

	// Blend the fog color and the lit color.
	litColor = lerp(litColor, gFogColor, fogLerp);

	// Common to take alpha from diffuse material.
	litColor.a = mt.Diffuse.a;

	return litColor;
}